import { EntityMetadata } from "../metadata/EntityMetadata";
import { ObjectLiteral } from "../common/ObjectLiteral";
import { Connection } from "../connection/Connection";
import { Subject } from "./subject/Subject";
import { SubjectCollection } from "./subject/SubjectCollection";
/**
 * To be able to execute persistence operations we need to load all entities from the database we need.
 * Loading should be efficient - we need to load entities in as few queries as possible + load as less data as we can.
 * This is how we determine which entities needs to be loaded from db:
 *
 * 1. example with cascade updates and inserts:
 *
 * [Y] - means "yes, we load"
 * [N] - means "no, we don't load"
 * in {} braces we specify what cascade options are set between relations
 *
 * if Post is new, author is not set in the post
 *
 * [Y] Post -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
 *   [Y] Author -> {all} // no because author is not set
 *     [Y] Photo -> {all} // no because author and its photo are not set
 *       [Y] Tag -> {all} // no because author and its photo and its tag are not set
 *
 * if Post is new, author is new (or anything else is new)
 * if Post is updated
 * if Post and/or Author are updated
 *
 * [Y] Post -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
 *   [Y] Author -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
 *     [Y] Photo -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
 *       [Y] Tag -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
 *
 * Here we load post, author, photo, tag to check if they are new or not to persist insert or update operation.
 * We load post, author, photo, tag only if they exist in the relation.
 * From these examples we can see that we always load entity relations when it has "update" or "insert" cascades.
 *
 * 2. example with cascade removes
 *
 * if entity is new its remove operations by cascades should not be executed
 * if entity is updated then values that are null or missing in array (not undefined!, undefined means skip - don't do anything) are treated as removed
 * if entity is removed then all its downside relations which has cascade remove should be removed
 *
 * Once we find removed entity - we load it, and every downside entity which has "remove" cascade set.
 *
 * At the end we have all entities we need to operate with.
 * Next step is to store all loaded entities to manipulate them efficiently.
 */
export declare class DatabaseEntityLoader<Entity extends ObjectLiteral> {
    protected connection: Connection;
    /**
     * If this gonna be reused then what to do with marked flags?
     * One of solution can be clone this object and reset all marked states for this persistence.
     * Or from reused just extract databaseEntities from their subjects? (looks better)
     */
    loadedSubjects: SubjectCollection;
    constructor(connection: Connection);
    persist(entity: Entity, metadata: EntityMetadata): Promise<void>;
    remove(entity: Entity, metadata: EntityMetadata): Promise<void>;
    /**
     * Extracts all entities that should be loaded from the database.
     * These are only entities which has insert and update cascades.
     * We can't add removed entities here, because to know which entity we need to load we entity id.
     * But in persist entity we don't have entity, thous don't have id.
     * Entity id of the removed entity is stored in the database entity.
     * That's why we first need to load all changed entities,
     * then extract ids of the removed entities from them,
     * and only then load removed entities by extracted ids.
     */
    protected populateSubjectsWithCascadeUpdateAndInsertEntities(entity: ObjectLiteral, metadata: EntityMetadata): void;
    /**
     */
    protected populateSubjectsWithCascadeRemoveEntities(entity: ObjectLiteral, metadata: EntityMetadata): void;
    /**
     * Loads database entities for all loaded subjects which does not have database entities set.
     */
    protected loadDatabaseEntities(): Promise<void>;
    /**
     *

    protected findCascadeInsertAndUpdateEntities(entity: ObjectLiteral, metadata: EntityMetadata): void {
        metadata
            .extractRelationValuesFromEntity(entity, metadata.relations)
            .forEach(([relation, value, valueMetadata]) => {

                // if there is a value in the relation and insert or update cascades are set - it means we must load entity
                if (relation.isEntityDefined(value) && (relation.isCascadeInsert || relation.isCascadeUpdate)) {

                    // add to the array of subjects to load only if there is no same entity there already
                    const subject = new Subject(valueMetadata, value); // todo: store relations inside to create correct order then?
                    subject.markedAsInserted = relation.isCascadeInsert;
                    subject.markedAsUpdated = relation.isCascadeUpdate;
                    this.loadedSubjects.push(subject);

                    // go recursively and find other entities to load by cascades in currently inserted entities
                    this.populateSubjectsWithCascadeUpdateAndInsertEntities(value, valueMetadata);
                }
            });
    }*/
    /**
     * We need to load removed entity when:
     *  - entity with relations is not new (this can be determined only after entity is loaded from db)
     *  - entity missing relation. When relation is simple
     *      - in the case of one-to-one owner (with join column) relation we need to load owner entity
     *      - in the case of one-to-one (without join column) relation we need to load inverse side entity
     *      - in the case of many-to-one relations we need to load entity itself
     *      - in the case of one-to-many relations we need to load entities by relation from inverse side
     *
     *  Before loading each entity we need to check in the loaded subjects - maybe it was already loaded.
     *
     *  BIG NOTE: objects are being removed by cascades not only when relation is removed, but also when
     *  relation is replaced (e.g. changed with different object).
     */
    protected findCascadeRemovedEntitiesToLoad(subject: Subject): Promise<void>;
    /**
     * Builds all junction insert operations used to insert new bind data into junction tables.
     */
    private buildInsertJunctionOperations();
    /**
     * Builds all junction remove operations used to remove bind data from junction tables.
     */
    private buildRemoveJunctionOperations();
}
