import { EntityMetadata } from "../metadata/EntityMetadata";
import { ObjectLiteral } from "../common/ObjectLiteral";
import { Connection } from "../connection/Connection";
import { QueryRunner } from "../query-runner/QueryRunner";
import { Subject } from "./subject/Subject";
/**
 * Manages entity persistence - insert, update and remove of entity.
 *
 *
 * Rules of updating by cascades.
 * Insert operation can lead to:
 *  - insert operations
 *  - update operations
 * Update operation can lead to:
 *  - insert operations
 *  - update operations
 *  - remove operations
 * Remove operation can lead to:
 *  - remove operation
 */
export declare class EntityPersister<Entity extends ObjectLiteral> {
    protected connection: Connection;
    protected metadata: EntityMetadata;
    protected queryRunner: QueryRunner;
    constructor(connection: Connection, metadata: EntityMetadata, queryRunner: QueryRunner);
    /**
     * To be able to execute persistence operations we need to load all entities from the database we need.
     * Loading should be efficient - we need to load entities in as few queries as possible + load as less data as we can.
     * This is how we determine which entities needs to be loaded from db:
     *
     * 1. example with cascade updates and inserts:
     *
     * [Y] - means "yes, we load"
     * [N] - means "no, we don't load"
     * in {} braces we specify what cascade options are set between relations
     *
     * if Post is new, author is not set in the post
     *
     * [Y] Post -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
     *   [Y] Author -> {all} // no because author is not set
     *     [Y] Photo -> {all} // no because author and its photo are not set
     *       [Y] Tag -> {all} // no because author and its photo and its tag are not set
     *
     * if Post is new, author is new (or anything else is new)
     * if Post is updated
     * if Post and/or Author are updated
     *
     * [Y] Post -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
     *   [Y] Author -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
     *     [Y] Photo -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
     *       [Y] Tag -> {all} // yes because of "update" and "insert" cascades, no because of "remove"
     *
     * Here we load post, author, photo, tag to check if they are new or not to persist insert or update operation.
     * We load post, author, photo, tag only if they exist in the relation.
     * From these examples we can see that we always load entity relations when it has "update" or "insert" cascades.
     *
     * 2. example with cascade removes
     *
     * if entity is new its remove operations by cascades should not be executed
     * if entity is updated then values that are null or missing in array (not undefined!, undefined means skip - don't do anything) are treated as removed
     * if entity is removed then all its downside relations which has cascade remove should be removed
     *
     * Once we find removed entity - we load it, and every downside entity which has "remove" cascade set.
     *
     * At the end we have all entities we need to operate with.
     * Next step is to store all loaded entities to manipulate them efficiently.

    async load(entity: Entity): Promise<void> {

        class LoadEntity {
            constructor(public target: Function|string,
                        public mixedId: any) {
            }
        }

        const loadMap: LoadEntity[] = [];

        const recursive = (entity: ObjectLiteral, metadata: EntityMetadata) => {
            metadata
                .extractRelationValuesFromEntity(entity, metadata.relations)
                .forEach(([relation, value, valueMetadata]) => {

                    if (relation.isEntityDefined(value) && (relation.isCascadeInsert || relation.isCascadeUpdate)) {
                        loadMap.push(new LoadEntity(valueMetadata.target, valueMetadata.getEntityIdMixedMap(value)));
                        recursive(value, valueMetadata);
                    }
                });
        };

        recursive(entity, this.metadata);

    }*/
    /**
     * Removes given entity from the database.
     */
    remove(entity: Entity): Promise<Entity>;
    /**
     * Persists given entity in the database.
     * Persistence is a complex process:
     * 1. flatten all persist entities
     * 2. load from the database all entities that has primary keys and might be updated
     */
    persist(entity: Entity): Promise<Entity>;
    /**
     * Post {
     *  Author {
     *   Photo {
     *   }
     *  }
     * }
     */
    /**
     * Remove operations differs a bit with other operations (insert and update).
     * We iterate throw persisted entities and its relations to find removed entities.
     * Once we find removed entities, we iterate throw database entities and find all relations
     * of the removed entity that has cascades and add them to list of removed entities too.
     *
     * @deprecated
     */
    private buildRemoveOperations(persistentSubject, persistentSubjects, dbSubjects);
    /**
     * @deprecated
     */
    private buildUpdateOperations(persistedSubject, persistentSubjects, dbSubjects);
    /**
     * Differentiate columns from the updated entity and entity stored in the database.
     *
     * @deprecated
     */
    private diffColumns(updatedSubject, dbSubject);
    /**
     * Difference columns of the owning one-to-one and many-to-one columns.
     *
     * @deprecated
     */
    private diffRelationalColumns(updatedSubject, dbSubject);
    /**
     * @deprecated
     */
    private buildInsertOperations(persistedSubject, persistentSubjects, dbSubjects);
    /**
     * When ORM loads dbEntity it uses joins to load all entity dependencies. However when dbEntity is newly persisted
     * to the db, but uses already exist in the db relational entities, those entities cannot be loaded, and will
     * absent in dbEntities. To fix it, we need to go throw all persistedEntities we have, find out those which have
     * ids, check if we did not load them yet and try to load them. This algorithm will make sure that all dbEntities
     * are loaded. Further it will help insert operations to work correctly.
     * @deprecated
     */
    protected findNotLoadedIds(persistedEntities: Subject[], dbEntities?: Subject[]): Promise<Subject[]>;
    /**
     * Extracts unique entities from given entity and all its downside relations.
     * @deprecated
     */
    protected flattenEntityRelationTree(entity: Entity, metadata: EntityMetadata): Subject[];
}
