"use strict";
const ColumnTypes_1 = require("../../metadata/types/ColumnTypes");
const DataTransformationUtils_1 = require("../../util/DataTransformationUtils");
/**
 */
class Subject {
    // -------------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------------
    constructor(metadata, entity, databaseEntity) {
        this.canBeInserted = false;
        this.canBeUpdated = false;
        this.mustBeRemoved = false;
        /**
         * List of relations which need to be unset.
         * This is used to update relation from inverse side.
         */
        this.relationUpdates = [];
        this.diffColumns = [];
        this.diffRelations = [];
        /**
         * Date when this entity is persisted.
         */
        this.date = new Date();
        this.junctionInserts = [];
        this.junctionRemoves = [];
        this.metadata = metadata;
        this.entity = entity; // todo: temporary
        this.databaseEntity = databaseEntity;
    }
    // -------------------------------------------------------------------------
    // Accessors
    // -------------------------------------------------------------------------
    get entityTarget() {
        return this.metadata.target;
    }
    /**
     * Returns readable / loggable name of the entity target.
     */
    get entityTargetName() {
        if (this.entityTarget instanceof Function) {
            if (this.entityTarget.name) {
                return this.entityTarget.name;
            }
        }
        return this.entityTarget;
    }
    get id() {
        return this.metadata.getEntityIdMap(this.entity);
    }
    get mixedId() {
        return this.metadata.getEntityIdMixedMap(this.entity);
    }
    get mustBeInserted() {
        return this.canBeInserted && !this.databaseEntity;
    }
    get mustBeUpdated() {
        return this.canBeUpdated && (this.diffColumns.length > 0 || this.diffRelations.length > 0);
    }
    get hasRelationUpdates() {
        return this.relationUpdates.length > 0;
    }
    get databaseEntity() {
        return this._databaseEntity;
    }
    set databaseEntity(databaseEntity) {
        this._databaseEntity = databaseEntity;
        if (this.entity && databaseEntity) {
            this.diffColumns = this.buildDiffColumns();
            this.diffRelations = this.buildDiffRelationalColumns();
        }
    }
    // -------------------------------------------------------------------------
    // Public Methods
    // -------------------------------------------------------------------------
    compareId(id) {
        return this.metadata.compareIds(this.id, id);
    }
    // -------------------------------------------------------------------------
    // Private Methods
    // -------------------------------------------------------------------------
    /**
     * Differentiate columns from the updated entity and entity stored in the database.
     */
    buildDiffColumns() {
        return this.metadata.allColumns.filter(column => {
            let entityValue = column.getEntityValue(this.entity);
            let databaseValue = column.getEntityValue(this.databaseEntity);
            if (entityValue !== null && entityValue !== undefined) {
                if (column.type === ColumnTypes_1.ColumnTypes.DATE) {
                    entityValue = DataTransformationUtils_1.DataTransformationUtils.mixedDateToDateString(entityValue);
                }
                else if (column.type === ColumnTypes_1.ColumnTypes.TIME) {
                    entityValue = DataTransformationUtils_1.DataTransformationUtils.mixedDateToTimeString(entityValue);
                }
                else if (column.type === ColumnTypes_1.ColumnTypes.DATETIME) {
                    if (column.loadInLocalTimezone) {
                        entityValue = DataTransformationUtils_1.DataTransformationUtils.mixedDateToDatetimeString(entityValue);
                        databaseValue = DataTransformationUtils_1.DataTransformationUtils.mixedDateToDatetimeString(databaseValue);
                    }
                    else {
                        entityValue = DataTransformationUtils_1.DataTransformationUtils.mixedDateToUtcDatetimeString(entityValue);
                        databaseValue = DataTransformationUtils_1.DataTransformationUtils.mixedDateToUtcDatetimeString(databaseValue);
                    }
                }
                else if (column.type === ColumnTypes_1.ColumnTypes.JSON) {
                    entityValue = JSON.stringify(entityValue);
                    if (databaseValue !== null && databaseValue !== undefined)
                        databaseValue = JSON.stringify(databaseValue);
                }
                else if (column.type === ColumnTypes_1.ColumnTypes.SIMPLE_ARRAY) {
                    entityValue = DataTransformationUtils_1.DataTransformationUtils.stringToSimpleArray(entityValue);
                    databaseValue = DataTransformationUtils_1.DataTransformationUtils.stringToSimpleArray(databaseValue);
                }
            }
            if (column.isVirtual ||
                column.isParentId ||
                column.isDiscriminator ||
                column.isUpdateDate ||
                column.isVersion ||
                column.isCreateDate ||
                this.entity[column.propertyName] === undefined ||
                entityValue === databaseValue)
                return false;
            // filter out "relational columns" only in the case if there is a relation object in entity
            if (!column.isInEmbedded && this.metadata.hasRelationWithDbName(column.propertyName)) {
                const relation = this.metadata.findRelationWithDbName(column.propertyName); // todo: why with dbName ?
                if (this.entity[relation.propertyName] !== null && this.entity[relation.propertyName] !== undefined)
                    return false;
            }
            return true;
        });
    }
    /**
     * Difference columns of the owning one-to-one and many-to-one columns.
     */
    buildDiffRelationalColumns() {
        return this.metadata.allRelations.filter(relation => {
            if (!relation.isManyToOne && !(relation.isOneToOne && relation.isOwning))
                return false;
            // here we cover two scenarios:
            // 1. related entity can be another entity which is natural way
            // 2. related entity can be entity id which is hacked way of updating entity
            // todo: what to do if there is a column with relationId? (cover this too?)
            const updatedEntityRelationId = this.entity[relation.propertyName] instanceof Object ?
                this.metadata.getEntityIdMixedMap(this.entity[relation.propertyName])
                : this.entity[relation.propertyName];
            // here because we have enabled RELATION_ID_VALUES option in the QueryBuilder when we loaded db entities
            // we have in the dbSubject only relationIds.
            // this allows us to compare relation id in the updated subject with id in the database.
            // note that we used relation.name instead of relation.propertyName because query builder with RELATION_ID_VALUES
            // returns values in the relation.name column, not relation.propertyName column
            const dbEntityRelationId = this.databaseEntity[relation.name];
            // todo: try to find if there is update by relation operation - we dont need to generate update relation operation for this
            // todo: if (updatesByRelations.find(operation => operation.targetEntity === this && operation.updatedRelation === relation))
            // todo:     return false;
            // we don't perform operation over undefined properties
            if (updatedEntityRelationId === undefined)
                return false;
            // if both are empty totally no need to do anything
            if ((updatedEntityRelationId === undefined || updatedEntityRelationId === null) &&
                (dbEntityRelationId === undefined || dbEntityRelationId === null))
                return false;
            // if relation ids aren't equal then we need to update them
            return updatedEntityRelationId !== dbEntityRelationId;
        });
    }
    /**
     * Gets id of the persisted entity.
     * If entity itself has an id then it simply returns it.
     * If entity does not have an id then it returns newly generated id.
     */
    getPersistedEntityIdMap() {
        return this.metadata.getDatabaseEntityIdMap(this.entity) ||
            this.metadata.createSimpleDatabaseIdMap(this.newlyGeneratedId);
    }
}
exports.Subject = Subject;

//# sourceMappingURL=Subject.js.map
